#include "ik.hpp"
#include "util.hpp"

using namespace glm;

constexpr float tolerance = 0.001;

const std::vector<glm::vec3> fabrik(const glm::vec3 t,
		const std::vector<glm::vec3> jpsIn, // joint positions
		const std::vector<float> jds // distances between each joint
		) {
	size_t N = jpsIn.size();
	assert(N == jds.size() + 1);
	std::vector<glm::vec3> jps = jpsIn;

	float dist = distance(jps[0], t);
	float totalLength = 0;
	for (int i = 0; i < N - 1; i++) totalLength += jds[i];
	if (dist > totalLength) { // target is unreachable
		for (int i = 0; i < N - 1; i++) {
			float r = distance(t, jps[i]);
			float lambda = jds[i] / r;
			jps[i + 1] = (1.f - lambda) * jps[i] + lambda * t;
		}
	} else { // target is reachable
		glm::vec3 b = jps[0];
		// distance between end effector and target
		float diff = distance(jps[N - 1], t);
		while (diff > tolerance) {
			// forward reaching
			jps[N - 1] = t; // set end effector to target
			for (int i = N - 2; i >= 0; i--) {
				float r = distance(jps[i + 1], jps[i]);
				float lambda = jds[i] / r;
				jps[i] = (1.f - lambda) * jps[i + 1] + lambda * jps[i];
			}

			// backward reaching
			jps[0] = b; // reset root to initial pos
			for (int i = 0; i < N - 1; i++) {
				float r = distance(jps[i + 1], jps[i]);
				float lambda = jds[i] / r;
				jps[i + 1] = (1 - lambda) * jps[i] + lambda * jps[i + 1];
			}

			float newDiff = distance(jps[N - 1], t);
			if (newDiff == diff) abort();
			diff = newDiff;
		}
	}
	return jps;
}

std::vector<Model::Node> allNodesTo(const Model::Node &from, const Model::Node &to) {
	if (from == to) return { to };
	assert(to.parent != nullptr);
	auto res = allNodesTo(from, *to.parent);
	res.push_back(to);
	return res;
}

mat4 getAbsTrans(const Model::Node &root, const Model::Node &n) {
	if (root.ai.mName == n.ai.mName)
		return mat4(1); //aiMatrixToMat4(n.ai.mTransformation);
	assert(n.parent != nullptr);
	return getAbsTrans(root, *n.parent) * aiMatrixToMat4(n.ai.mTransformation);
}

vec3 extractPos(mat4 trans) {
	return vec3(trans[3]);
}

mat4 absoluteToModelSpace(const Model::Node &root, const Model::Node &n, mat4 m) {
	const Model::Node *parent = &n;
	glm::mat4 res = m;
	std::vector<mat4> trans;
	while (&parent->ai != &root.ai) {
		trans.push_back(inverse(aiMatrixToMat4(parent->ai.mTransformation)));
		parent = parent->parent;
	}
	while (!trans.empty()) { res = trans.back() * res; trans.pop_back(); }
	return res;
}

// Given points u and v on a sphere, calculate the transformation matrix
// to bring u -> v
// from https://math.stackexchange.com/a/2765250
mat4 getRotationToPoint(vec3 u, vec3 v, float dist) {
	if (distance(u, v) < 0.00001) return mat4(1);
	vec3 n = cross(u, v) / length(cross(u, v));
	vec3 t = cross(n, u);
	float alpha = atan2(dot(v, t), dot(v, u));
	mat4 T = mat4(mat3(u, t, n));
	mat4 R = rotate(mat4(1), alpha, {0, 0, 1}); // rotation in z axis
	mat4 res = T * R * inverse(T);
	return res;
}

// Normalizes a transformation matrix to not have any scale factor
inline mat4 normalizeScale(mat4 m) {
	for (int i = 0; i < 3; i++)
		m[i] = normalize(m[i]);
	return m;
}

float d = 0;

// Target is world position
void inverseKinematics(Model::Node &start, Model::Node &end, vec3 target) {
	std::vector<Model::Node> chain = allNodesTo(start, end);
	assert(!chain.empty());

	// Calculate the world root
	const Model::Node &root = start.getRoot();

	// Work out the positions and distances
	std::vector<vec3> positions(chain.size()); std::vector<float> distances(chain.size() - 1);
	for (size_t i = 0; i < chain.size(); i++) {
		mat4 absTrans = getAbsTrans(root, chain[i]);
		positions[i] = extractPos(absTrans);
		if (i > 0)
			distances[i - 1] = distance(positions[i], positions[i - 1]);
	}

	// Do the actual IK part
	auto newPositions = fabrik(target, positions, distances);
	
	// Move all the nodes so that they are in the correct positions
	// Don't need to move the root node - it's already in place
	for (size_t i = 1; i < chain.size(); i++) {
		auto node = chain[i];
		mat4 absTrans = getAbsTrans(root, node);
		absTrans[3] = vec4(newPositions[i], absTrans[3][3]); // update position in transform
		
		mat4 relTrans = absoluteToModelSpace(root, *node.parent, absTrans);
		node.ai.mTransformation = mat4ToaiMatrix(relTrans);
	}

	// Now rotate all the nodes so that they point towards each other
	// Don't need to rotate the last node - it has nothing to point towards
	// FIXME: Despite normalizeScale, this is still numerically unstable
	// and the transformation scales over time!!!
	for (size_t i = 0; i < chain.size() - 1; i++) {
		auto node = chain[i]; auto nextNode = chain[i + 1];
		mat4 oldTrans = aiMatrixToMat4(node.ai.mTransformation);
		vec3 nextNodePos = extractPos(aiMatrixToMat4(nextNode.ai.mTransformation));

		vec3 up = {0, 1, 0};
		vec3 dir = -normalize(nextNodePos);

		vec3 v = cross(up, dir);
		mat3 sscpm = mat3(0, -v[2], v[1],
						  v[2], 0, -v[0],
						  -v[1], v[0], 0);
		mat4 rot = mat3(1) + sscpm + sscpm * sscpm * (1.f / 1.f + dot(up, dir));

		// very important that we normalize the scale
		// otherwise we end up with gradually growing models
		// due to numerical instabaility
		rot = normalizeScale(rot);
		node.ai.mTransformation = mat4ToaiMatrix(oldTrans * rot);

		for (auto child: node.getChildren()) {
			child->ai.mTransformation = mat4ToaiMatrix(normalizeScale(inverse(rot)) * aiMatrixToMat4(child->ai.mTransformation));
		}

	}

}

