#include "model.hpp"
#include <iostream>
#include <assimp/quaternion.h>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/closest_point.hpp>
#include "util.hpp"

Model::Mesh::Mesh(const aiMesh *aiMesh, GLuint progId) : progId(progId), ai(*aiMesh) {
	std::vector<glm::vec3> vertices, normals, tangents, bitangents;
	std::vector<glm::vec2> texCoords;

	for (int i = 0; i < aiMesh->mNumVertices; i++) {
		if (aiMesh->HasPositions()) {
			aiVector3D v = aiMesh->mVertices[i];
			vertices.push_back(glm::vec3(v.x, v.y, v.z));
		}
		if (aiMesh->HasNormals()) {
			aiVector3D v = aiMesh->mNormals[i];
			normals.push_back(glm::vec3(v.x, v.y, v.z));
		} else {
			std::cerr << "Missing normals" << std::endl;
			abort();
		}
		// check for texture coord set 0
		if (aiMesh->HasTextureCoords(0)) {
			const aiVector3D v = aiMesh->mTextureCoords[0][i];
			texCoords.push_back(glm::vec2(v.x, v.y));
		} else {
			texCoords.push_back(glm::vec2(0));
		}
		materialIndex = aiMesh->mMaterialIndex;
	}

	std::vector<GLuint> indices;

	for (int i = 0; i < aiMesh->mNumFaces; i++) {
		const aiFace &face = aiMesh->mFaces[i];
		if(face.mNumIndices == 3) {
			indices.push_back(face.mIndices[0]);
			indices.push_back(face.mIndices[1]);
		   	indices.push_back(face.mIndices[2]);
		}
	} 
	
	numIndices = indices.size();
	
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);
	
	glGenBuffers(6, vbos);
	GLuint posVbo = vbos[0], normalVbo = vbos[1], texCoordVbo = vbos[2], indicesVbo = vbos[3];
	GLuint boneVbo = vbos[4];
	
	GLuint posLoc = glGetAttribLocation(progId, "pos");
	glBindBuffer(GL_ARRAY_BUFFER, posVbo);
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec3), &vertices[0], GL_STATIC_DRAW);
	glEnableVertexAttribArray(posLoc);
	glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	GLuint normalLoc = glGetAttribLocation(progId, "unscaledNormal");
	glBindBuffer(GL_ARRAY_BUFFER, normalVbo);
	glBufferData(GL_ARRAY_BUFFER, normals.size() * sizeof(glm::vec3), &normals[0], GL_STATIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	GLuint texCoordLoc = glGetAttribLocation(progId, "vTexCoord");
	glBindBuffer(GL_ARRAY_BUFFER, texCoordVbo);
	glBufferData(GL_ARRAY_BUFFER, texCoords.size() * sizeof(glm::vec2), &texCoords[0], GL_STATIC_DRAW);
	glEnableVertexAttribArray(texCoordLoc);
	glVertexAttribPointer(texCoordLoc, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indicesVbo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint), &indices[0], GL_STATIC_DRAW);

	// bones
	std::vector<VertBones> vertBones(aiMesh->mNumVertices);

	std::map<unsigned int, std::vector<std::pair<unsigned int, float>>> boneWeightMap;

	for (unsigned int i = 0; i < aiMesh->mNumBones; i++) {
		aiBone *aiBone = aiMesh->mBones[i];

		boneMap[std::string(aiBone->mName.C_Str())] = std::pair(i + 1, aiBone);

		for (int j = 0; j < aiBone->mNumWeights; j++) {
			aiVertexWeight vw = aiBone->mWeights[j];

			if (!boneWeightMap.count(vw.mVertexId))
				boneWeightMap[vw.mVertexId] = std::vector<std::pair<unsigned int, float>>();
			boneWeightMap[vw.mVertexId].push_back(std::pair(i + 1, vw.mWeight));
		}
	}

	for (auto pair: boneWeightMap) {
		unsigned int vertexId = pair.first;
		for (int i = 0; i < pair.second.size() && i < 4; i++) {
			unsigned int boneId = pair.second[i].first;
			float weight = pair.second[i].second;
			vertBones[vertexId].ids[i] = boneId;
		   	vertBones[vertexId].weights[i] = weight;
		}
	}
	
	glBindBuffer(GL_ARRAY_BUFFER, boneVbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(VertBones) * vertBones.size(), &vertBones[0], GL_STATIC_DRAW);

	GLuint boneIdLoc = glGetAttribLocation(progId, "boneIds");
	glEnableVertexAttribArray(boneIdLoc);
	glVertexAttribIPointer(boneIdLoc, 4, GL_INT, sizeof(VertBones), 0);

	GLuint boneWeightLoc = glGetAttribLocation(progId, "boneWeights");
	glEnableVertexAttribArray(boneWeightLoc);
	glVertexAttribPointer(boneWeightLoc, 4, GL_FLOAT, GL_FALSE, sizeof(VertBones), (const GLvoid *)sizeof(VertBones::ids));
}

void Model::Mesh::updatePosBuffer() const {
	GLuint posLoc = glGetAttribLocation(progId, "pos");
	GLuint posVbo = vbos[0];
	glBindBuffer(GL_ARRAY_BUFFER, posVbo);
	glBufferData(GL_ARRAY_BUFFER, ai.mNumVertices * sizeof(aiVector3D), ai.mVertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(posLoc);
	glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
}

Model::Node::Node(aiNode &node, GLuint progId, AnimMap *am, std::set<std::string> allBones, Node *p): ai(node), parent(p), progId(progId), animMap(am), isBone(allBones.count(std::string(node.mName.C_Str())) > 0) {
	for (int i = 0; i < node.mNumMeshes; i++) {
		meshIndices.push_back(node.mMeshes[i]);
	}
	for (int i = 0; i < node.mNumChildren; i++) {
		aiNode *child = node.mChildren[i];
		children.push_back(new Node(*child, progId, am, allBones, this));
	}
}

glm::mat4 lerpPosition(const aiNodeAnim *anim, const float tick) {
	if (anim->mNumPositionKeys == 0) return glm::mat4(1.f);

	int yIndex = -1;
	for (int i = 0; i < anim->mNumPositionKeys; i++) {
		aiVectorKey vk = anim->mPositionKeys[i];
		if (vk.mTime > tick) {
			yIndex = i;
			break;
		}
	}
	aiVector3D lerpPos;
	if (yIndex == 0) {
		lerpPos = anim->mPositionKeys[0].mValue;
	} else if (yIndex == -1) {
		lerpPos = anim->mPositionKeys[anim->mNumPositionKeys - 1].mValue;
	} else {
		auto X = anim->mPositionKeys[yIndex - 1];
		auto Y = anim->mPositionKeys[yIndex];

		lerpPos = (X.mValue * (float)(Y.mTime - tick) + Y.mValue * (float)(tick - X.mTime)) / (float)(Y.mTime - X.mTime);
	}
	aiMatrix4x4 result;
	aiMatrix4x4::Translation(lerpPos, result);
	return aiMatrixToMat4(result);
}

glm::mat4 lerpRotation(const aiNodeAnim *anim, const float tick) {
	int yIndex = -1;
	for (int i = 0; i < anim->mNumRotationKeys; i++) {
		aiQuatKey vk = anim->mRotationKeys[i];
		if (vk.mTime > tick) {
			yIndex = i;
			break;
		}
	}

	aiQuaternion result;
	if (yIndex < 1) {
		result = anim->mRotationKeys[0].mValue;
	} else if (yIndex == -1) {
		result = anim->mRotationKeys[anim->mNumRotationKeys - 1].mValue;
	} else {

		auto X = anim->mRotationKeys[yIndex - 1];
		auto Y = anim->mRotationKeys[yIndex];

		float mix = (tick - X.mTime) / (Y.mTime - X.mTime);

		aiQuaternion::Interpolate(result, X.mValue, Y.mValue, mix);

	}
	return aiMatrixToMat4(aiMatrix4x4(result.GetMatrix()));
}

glm::mat4 lerpScaling(const aiNodeAnim *anim, const float tick) {
	int yIndex = -1;
	for (int i = 0; i < anim->mNumScalingKeys; i++) {
		aiVectorKey vk = anim->mScalingKeys[i];
		if (vk.mTime > tick) {
			yIndex = i;
			break;
		}
	}

	aiVector3D lerpPos;
	if (yIndex < 1) {
		lerpPos = anim->mScalingKeys[0].mValue;
	} else {
		auto X = anim->mScalingKeys[yIndex - 1];
		auto Y = anim->mScalingKeys[yIndex];

		lerpPos = (X.mValue * (float)(Y.mTime - tick) + Y.mValue * (float)(tick - X.mTime)) / (float)(Y.mTime - X.mTime);
	}
	aiMatrix4x4 result;
	aiMatrix4x4::Scaling(lerpPos, result);
	return aiMatrixToMat4(result);
}

glm::mat4 Model::Node::totalTrans(const glm::mat4 parentTrans, const float tick) const {
	glm::mat4 aiTrans = aiMatrixToMat4(ai.mTransformation);
	if (animMap->count(std::string(ai.mName.C_Str()))) {
		for (const Animation anim: animMap->at(std::string(ai.mName.C_Str()))) {
			// animations are *absolute*
			// they replace aiNode.mTransformation!!
			aiTrans = glm::mat4(1);
			float t = fmod(tick, anim.duration);
			for (const aiNodeAnim *nodeAnim: anim.nodeAnims) {
				aiTrans *= lerpPosition(nodeAnim, t);
				aiTrans *= lerpRotation(nodeAnim, t);
				aiTrans *= lerpScaling(nodeAnim, t);
			}
		}
	}

	glm::mat4 m = parentTrans * aiTrans * transform;
	return m;
}

const Model::Node &Model::Node::getRoot() const {
	const Model::Node *rootPtr = this;
	while (rootPtr->parent != nullptr)
		rootPtr = rootPtr->parent;
	const Model::Node &root = *rootPtr;
	return root;
}

void Model::Node::draw(	const std::vector<Mesh> &meshes,
						const std::vector<Material> &materials,
						const Skybox skybox,
						const float tick,
						const BoneTransforms &boneTransforms,
						glm::mat4 parentTrans = glm::mat4(1)) const {

	GLuint modelLoc = glGetUniformLocation(progId, "model");
	glm::mat4 m = totalTrans(parentTrans, tick);

#ifdef DEBUG_NODES
	if (isBone)
		drawDebugNode(m, {0, 0.5, 1, 1});
	else
		drawDebugNode(m);
#endif

	for (unsigned int i: meshIndices) {
		const Mesh &mesh = meshes[i];
		glBindVertexArray(mesh.vao);

		// bones
		std::vector<glm::mat4> idBones(17, glm::mat4(1.f));
		glUniformMatrix4fv(glGetUniformLocation(progId, "bones"), 17, GL_FALSE, glm::value_ptr(idBones[0]));

		// bonemap: map from bone nodes to bone ids and aiBones
		for (auto pair: mesh.boneMap) {

			std::string boneName = pair.first;

			unsigned int boneId = pair.second.first;
			aiBone *bone = pair.second.second;
			// This is actually an inverse-bind matrix
			// i.e. transforms bone space -> mesh space
			// so no need to inverse again!
			// https://github.com/assimp/assimp/pull/1803/files
			glm::mat4 boneOffset = aiMatrixToMat4(bone->mOffsetMatrix);

			if (!boneTransforms.count(boneName)) abort();
			glm::mat4 boneTrans = boneTransforms.at(boneName);

			boneTrans = boneTrans * boneOffset;

			std::string boneLocStr = "bones[" + std::to_string(boneId) + "]";
			GLuint boneLoc = glGetUniformLocation(progId, boneLocStr.c_str());
			glUniformMatrix4fv(boneLoc, 1, GL_FALSE, glm::value_ptr(boneTrans));
		}

		Material material = materials[mesh.materialIndex];
		material.bind();

		glUniform1i(glGetUniformLocation(progId, "irradianceMap"), 4);
		glActiveTexture(GL_TEXTURE4);
		glBindTexture(GL_TEXTURE_CUBE_MAP, skybox.getIrradianceMap());

		glUniform1i(glGetUniformLocation(progId, "prefilterMap"), 5);
		glActiveTexture(GL_TEXTURE5);
		glBindTexture(GL_TEXTURE_CUBE_MAP, skybox.getPrefilterMap());

		glUniform1i(glGetUniformLocation(progId, "brdfMap"), 6);
		glActiveTexture(GL_TEXTURE6);
		glBindTexture(GL_TEXTURE_2D, skybox.getBRDFMap());
		
		glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(m));

		glDrawElements(GL_TRIANGLES, mesh.numIndices, GL_UNSIGNED_INT, 0);
	}
	for (Node *child: children) child->draw(meshes, materials, skybox, tick, boneTransforms, m);
}

void printHierarchy(aiNode *n, int indent = 0) {
	for (int i = 0; i < indent; i++)
		fprintf(stderr, "\t");
	fprintf(stderr,"%s\n", n->mName.C_Str());
	printMatrix4x4(n->mTransformation);
	for (int i = 0; i < n->mNumChildren; i++)
		printHierarchy(n->mChildren[i], indent + 1);
}

Model::Model(const aiScene *scene, Program p): program(p) {
	glUseProgram(p.progId);

	std::set<std::string> allBones;
	for (int i = 0; i < scene->mNumMeshes; i++) {
		const aiMesh *mesh = scene->mMeshes[i];
		meshes.push_back(Mesh(mesh, p.progId));
		for (int j = 0; j < mesh->mNumBones; j++)
			allBones.insert(std::string(mesh->mBones[j]->mName.C_Str()));
	}

	for (unsigned int i = 0; i < scene->mNumMaterials; i++) {
		const aiMaterial &material = *scene->mMaterials[i];
		materials.push_back(Material(material, *scene, p.progId));
	}

	for (int i = 0; i < scene->mNumAnimations; i++) {
		const aiAnimation *aiAnim = scene->mAnimations[i];

		std::map<std::string, std::vector<const aiNodeAnim*>> nodeAnims;

		for (int j = 0; j < aiAnim->mNumChannels; j++) {
			const aiNodeAnim *nodeAnim = aiAnim->mChannels[j];
			std::string nodeName = std::string(nodeAnim->mNodeName.C_Str());
			
			if (!nodeAnims.count(nodeName)) nodeAnims[nodeName] = std::vector<const aiNodeAnim*>();

			nodeAnims[nodeName].push_back(nodeAnim);
		}

		for (std::pair<std::string, std::vector<const aiNodeAnim*>> pair: nodeAnims) {
			std::string nodeName = pair.first;

			if (!animMap.count(nodeName)) animMap[nodeName] = std::vector<const Animation>();
			animMap[nodeName].push_back({ aiAnim->mDuration, pair.second });
		}
	}

	root = new Node(*(scene->mRootNode), p.progId, &animMap, allBones, nullptr);
}


std::map<std::string, glm::mat4> Model::calcBoneTransforms(const Node &n, const float tick, const std::set<std::string> bones, const glm::mat4 parentTrans = glm::mat4(1)) const {
	std::string name = std::string(n.ai.mName.C_Str());

	glm::mat4 m = n.totalTrans(parentTrans, tick);

	BoneTransforms res;
	if (bones.count(name) > 0)
		res[std::string(n.ai.mName.C_Str())] = m; // take part in hierarchy
	else
		m = glm::mat4(1); // ignore this node transformation
	for (const auto child: n.getChildren())
		res.merge(calcBoneTransforms(*child, tick, bones, m));
	return res;
}

void Model::draw(Skybox skybox, const float tick) const {
	glUseProgram(program.progId);

	std::set<std::string> bones;
	for (auto m: this->meshes) {
		for (auto b: m.boneMap) {
			bones.insert(b.first);
		}
	}
	auto boneTransforms = calcBoneTransforms(*root, tick, bones);

	root->draw(meshes, materials, skybox, tick, boneTransforms);
}

Model::Node* Model::find(const std::string &name) const {
	return find(aiString(name));
}

Model::Node* Model::find(const aiString name) const {
	const aiNode *node = root->ai.FindNode(name);
	Model::Node* res = root->findNode(*node);
	return res;
}

Model::Node* Model::Node::findNode(const aiNode &aiNode) {
	if (&ai == &aiNode) return this;
	for (Model::Node *child: children) {
		Model::Node *res = child->findNode(aiNode);
		if (res) return res;
	}
	return nullptr;
}

bool Model::Node::operator==(const Model::Node &rhs) const {
	return &ai == &rhs.ai;
}

// Returns closest vertex in world space and distance
// a and b define the line in 3d space
Model::VertexLookup Model::closestVertex(Model::Node &n, glm::vec3 a, glm::vec3 b, glm::mat4 parentTrans) const {
	Model::VertexLookup closest;
	closest.distance = FLT_MAX;

	for (int i = 0; i < n.ai.mNumMeshes; i++) {
		int meshIdx = n.ai.mMeshes[i];
		const aiMesh &mesh = meshes[meshIdx].ai;

		for (int j = 0; j < mesh.mNumVertices; j++) {
			if (mesh.HasNormals()) {
				auto n = aiVector3DToVec3(mesh.mNormals[j]);
				if (glm::dot(n, glm::normalize(b - a)) > 0)
					continue;
			}
			glm::vec4 vPos = glm::vec4(aiVector3DToVec3(mesh.mVertices[j]), 1);
			// Move from model space -> world space
			vPos = parentTrans * aiMatrixToMat4(n.ai.mTransformation) * vPos;
			float dist = glm::distance(glm::vec3(vPos),
							glm::closestPointOnLine(glm::vec3(vPos), a, b));
			if (dist < closest.distance) {
				closest.pos = glm::vec3(vPos);
				closest.distance = dist;
				closest.meshIdx = i;
				closest.vertIdx = j;
			}
		}
	}
	
	for (auto child: n.getChildren()) {
		auto childRes = closestVertex(*child, a, b, parentTrans * aiMatrixToMat4(n.ai.mTransformation));
		if (childRes.distance < closest.distance) 
			closest = childRes;
	}

	return closest;
}
